@@ -76,8 +76,8 @@ void bisexec_version(const char *const *argv) {
 	 BASE_MAGIC,
 	 MAX_ALLOW_FD, MAX_GENERAL_STRING,
 	 MAX_GIDS, MAX_OVERRIDE_LEN,
-	 MAX_ARGSDEFVAR, ERRMSG_RESERVE_ERRNO,
-	 MAX_INCLUDE_NEST, MAX_ERRMSG_LEN);
+	 MAX_ARGSDEFVAR, MAX_ERRMSG_LEN,
+	 MAX_INCLUDE_NEST, ERRMSG_RESERVE_ERRNO);
   for (i=0, p=protocolchecksumversion; i<sizeof(protocolchecksumversion); i++, p++)
     printf("%02x",*p);
   printf("\n"
